package java.time;

import java.io.ObjectInput;
import java.io.IOException;
import java.io.ObjectOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.TemporalQueries;
import java.time.temporal.TemporalQuery;
import java.time.temporal.TemporalAmount;
import java.time.temporal.ValueRange;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalUnit;
import java.time.temporal.ChronoField;
import java.time.temporal.TemporalField;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAccessor;
import java.util.Objects;
import java.io.Serializable;
import java.time.temporal.TemporalAdjuster;
import java.time.temporal.Temporal;

public final class OffsetTime implements Temporal, TemporalAdjuster, Comparable<OffsetTime>, Serializable
{
    public static final OffsetTime MIN;
    public static final OffsetTime MAX;
    private static final long serialVersionUID = 7264499704384272492L;
    private final LocalTime time;
    private final ZoneOffset offset;
    
    public static OffsetTime now() {
        return now(Clock.systemDefaultZone());
    }
    
    public static OffsetTime now(ZoneId zoneId) {
        return now(Clock.system(zoneId));
    }
    
    public static OffsetTime now(Clock obj) {
        Objects.requireNonNull(obj, "clock");
        Instant instant = obj.instant();
        return ofInstant(instant, obj.getZone().getRules().getOffset(instant));
    }
    
    public static OffsetTime of(LocalTime localTime, ZoneOffset zoneOffset) {
        return new OffsetTime(localTime, zoneOffset);
    }
    
    public static OffsetTime of(int n, int n2, int n3, int n4, ZoneOffset zoneOffset) {
        return new OffsetTime(LocalTime.of(n, n2, n3, n4), zoneOffset);
    }
    
    public static OffsetTime ofInstant(Instant instant, ZoneId obj) {
        Objects.requireNonNull(instant, "instant");
        Objects.requireNonNull(obj, "zone");
        ZoneOffset offset = obj.getRules().getOffset(instant);
        return new OffsetTime(LocalTime.ofNanoOfDay((long)(int)Math.floorMod(instant.getEpochSecond() + (long)offset.getTotalSeconds(), 86400L) * 1000000000L + (long)instant.getNano()), offset);
    }
    
    public static OffsetTime from(TemporalAccessor obj) {
        if ((obj instanceof OffsetTime)) {
            return (OffsetTime)obj;
        }
        try {
            return new OffsetTime(LocalTime.from(obj), ZoneOffset.from(obj));
        }
        catch (DateTimeException ex) {
            throw new DateTimeException("Unable to obtain OffsetTime from TemporalAccessor: " + obj + " of type " + obj.getClass().getName(), ex);
        }
    }
    
    public static OffsetTime parse(CharSequence charSequence) {
        return parse(charSequence, DateTimeFormatter.ISO_OFFSET_TIME);
    }
    
    public static OffsetTime parse(CharSequence text, DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return (OffsetTime)obj.parse(text, OffsetTime::from);
    }
    
    private OffsetTime(LocalTime obj, ZoneOffset obj2) {
        this.time = (LocalTime)Objects.requireNonNull(obj, "time");
        this.offset = (ZoneOffset)Objects.requireNonNull(obj2, "offset");
    }
    
    private OffsetTime with(LocalTime localTime, ZoneOffset zoneOffset) {
        if (this.time == localTime && this.offset.equals(zoneOffset)) {
            return this;
        }
        return new OffsetTime(localTime, zoneOffset);
    }
    
    @Override
    public boolean isSupported(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            return temporalField.isTimeBased() || temporalField == ChronoField.OFFSET_SECONDS;
        }
        return temporalField != null && temporalField.isSupportedBy(this);
    }
    
    @Override
    public boolean isSupported(TemporalUnit temporalUnit) {
        if ((temporalUnit instanceof ChronoUnit)) {
            return temporalUnit.isTimeBased();
        }
        return temporalUnit != null && temporalUnit.isSupportedBy(this);
    }
    
    @Override
    public ValueRange range(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            if (temporalField == ChronoField.OFFSET_SECONDS) {
                return temporalField.range();
            }
            return this.time.range(temporalField);
        }
        else {
            return temporalField.rangeRefinedBy(this);
        }
    }
    
    @Override
    public int get(TemporalField field) {
        return super.get(field);
    }
    
    @Override
    public long getLong(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            if (temporalField == ChronoField.OFFSET_SECONDS) {
                return (long)this.offset.getTotalSeconds();
            }
            return this.time.getLong(temporalField);
        }
        else {
            return temporalField.getFrom(this);
        }
    }
    
    public ZoneOffset getOffset() {
        return this.offset;
    }
    
    public OffsetTime withOffsetSameLocal(ZoneOffset zoneOffset) {
        return (zoneOffset != null && zoneOffset.equals(this.offset)) ? this : new OffsetTime(this.time, zoneOffset);
    }
    
    public OffsetTime withOffsetSameInstant(ZoneOffset zoneOffset) {
        if (zoneOffset.equals(this.offset)) {
            return this;
        }
        return new OffsetTime(this.time.plusSeconds((long)(zoneOffset.getTotalSeconds() - this.offset.getTotalSeconds())), zoneOffset);
    }
    
    public LocalTime toLocalTime() {
        return this.time;
    }
    
    public int getHour() {
        return this.time.getHour();
    }
    
    public int getMinute() {
        return this.time.getMinute();
    }
    
    public int getSecond() {
        return this.time.getSecond();
    }
    
    public int getNano() {
        return this.time.getNano();
    }
    
    @Override
    public OffsetTime with(TemporalAdjuster temporalAdjuster) {
        if ((temporalAdjuster instanceof LocalTime)) {
            return this.with((LocalTime)temporalAdjuster, this.offset);
        }
        if ((temporalAdjuster instanceof ZoneOffset)) {
            return this.with(this.time, (ZoneOffset)temporalAdjuster);
        }
        if ((temporalAdjuster instanceof OffsetTime)) {
            return (OffsetTime)temporalAdjuster;
        }
        return (OffsetTime)temporalAdjuster.adjustInto(this);
    }
    
    @Override
    public OffsetTime with(TemporalField temporalField, long value) {
        if ((temporalField instanceof ChronoField)) {
            if (temporalField == ChronoField.OFFSET_SECONDS) {
                return this.with(this.time, ZoneOffset.ofTotalSeconds(((ChronoField)temporalField).checkValidIntValue(value)));
            }
            return this.with(this.time.with(temporalField, value), this.offset);
        }
        else {
            return (OffsetTime)temporalField.adjustInto(this, value);
        }
    }
    
    public OffsetTime withHour(int n) {
        return this.with(this.time.withHour(n), this.offset);
    }
    
    public OffsetTime withMinute(int n) {
        return this.with(this.time.withMinute(n), this.offset);
    }
    
    public OffsetTime withSecond(int n) {
        return this.with(this.time.withSecond(n), this.offset);
    }
    
    public OffsetTime withNano(int n) {
        return this.with(this.time.withNano(n), this.offset);
    }
    
    public OffsetTime truncatedTo(TemporalUnit temporalUnit) {
        return this.with(this.time.truncatedTo(temporalUnit), this.offset);
    }
    
    @Override
    public OffsetTime plus(TemporalAmount temporalAmount) {
        return (OffsetTime)temporalAmount.addTo(this);
    }
    
    @Override
    public OffsetTime plus(long n, TemporalUnit temporalUnit) {
        if ((temporalUnit instanceof ChronoUnit)) {
            return this.with(this.time.plus(n, temporalUnit), this.offset);
        }
        return (OffsetTime)temporalUnit.addTo(this, n);
    }
    
    public OffsetTime plusHours(long n) {
        return this.with(this.time.plusHours(n), this.offset);
    }
    
    public OffsetTime plusMinutes(long n) {
        return this.with(this.time.plusMinutes(n), this.offset);
    }
    
    public OffsetTime plusSeconds(long n) {
        return this.with(this.time.plusSeconds(n), this.offset);
    }
    
    public OffsetTime plusNanos(long n) {
        return this.with(this.time.plusNanos(n), this.offset);
    }
    
    @Override
    public OffsetTime minus(TemporalAmount temporalAmount) {
        return (OffsetTime)temporalAmount.subtractFrom(this);
    }
    
    @Override
    public OffsetTime minus(long n, TemporalUnit temporalUnit) {
        return (n == Long.MIN_VALUE) ? this.plus(Long.MAX_VALUE, temporalUnit).plus(1L, temporalUnit) : this.plus(-n, temporalUnit);
    }
    
    public OffsetTime minusHours(long n) {
        return this.with(this.time.minusHours(n), this.offset);
    }
    
    public OffsetTime minusMinutes(long n) {
        return this.with(this.time.minusMinutes(n), this.offset);
    }
    
    public OffsetTime minusSeconds(long n) {
        return this.with(this.time.minusSeconds(n), this.offset);
    }
    
    public OffsetTime minusNanos(long n) {
        return this.with(this.time.minusNanos(n), this.offset);
    }
    
    @Override
    public <R> R query(TemporalQuery<R> temporalQuery) {
        if (temporalQuery == TemporalQueries.offset() || temporalQuery == TemporalQueries.zone()) {
            return this.offset;
        }
        if ((temporalQuery == TemporalQueries.zoneId() | temporalQuery == TemporalQueries.chronology()) || temporalQuery == TemporalQueries.localDate()) {
            return null;
        }
        if (temporalQuery == TemporalQueries.localTime()) {
            return this.time;
        }
        if (temporalQuery == TemporalQueries.precision()) {
            return ChronoUnit.NANOS;
        }
        return temporalQuery.queryFrom(this);
    }
    
    @Override
    public Temporal adjustInto(Temporal temporal) {
        return temporal.with(ChronoField.NANO_OF_DAY, this.time.toNanoOfDay()).with(ChronoField.OFFSET_SECONDS, (long)this.offset.getTotalSeconds());
    }
    
    @Override
    public long until(Temporal temporal, TemporalUnit obj) {
        OffsetTime from = from(temporal);
        if ((obj instanceof ChronoUnit)) {
            long n = from.toEpochNano() - this.toEpochNano();
            switch ((ChronoUnit)obj) {
                case NANOS: {
                    return n;
                }
                case MICROS: {
                    return n / 1000L;
                }
                case MILLIS: {
                    return n / 1000000L;
                }
                case SECONDS: {
                    return n / 1000000000L;
                }
                case MINUTES: {
                    return n / 60000000000L;
                }
                case HOURS: {
                    return n / 3600000000000L;
                }
                case HALF_DAYS: {
                    return n / 43200000000000L;
                }
                default: {
                    throw new UnsupportedTemporalTypeException("Unsupported unit: " + obj);
                }
            }
        }
        else {
            return obj.between(this, from);
        }
    }
    
    public String format(DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return obj.format(this);
    }
    
    public OffsetDateTime atDate(LocalDate localDate) {
        return OffsetDateTime.of(localDate, this.time, this.offset);
    }
    
    private long toEpochNano() {
        return this.time.toNanoOfDay() - (long)this.offset.getTotalSeconds() * 1000000000L;
    }
    
    @Override
    public int compareTo(OffsetTime offsetTime) {
        if (this.offset.equals(offsetTime.offset)) {
            return this.time.compareTo(offsetTime.time);
        }
        int n = Long.compare(this.toEpochNano(), offsetTime.toEpochNano());
        if (n == 0) {
            n = this.time.compareTo(offsetTime.time);
        }
        return n;
    }
    
    public boolean isAfter(OffsetTime offsetTime) {
        return this.toEpochNano() > offsetTime.toEpochNano();
    }
    
    public boolean isBefore(OffsetTime offsetTime) {
        return this.toEpochNano() < offsetTime.toEpochNano();
    }
    
    public boolean isEqual(OffsetTime offsetTime) {
        return this.toEpochNano() == offsetTime.toEpochNano();
    }
    
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if ((o instanceof OffsetTime)) {
            OffsetTime offsetTime = (OffsetTime)o;
            return this.time.equals(offsetTime.time) && this.offset.equals(offsetTime.offset);
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return this.time.hashCode() ^ this.offset.hashCode();
    }
    
    @Override
    public String toString() {
        return this.time.toString() + this.offset.toString();
    }
    
    private Object writeReplace() {
        return new Ser((byte)9, this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    void writeExternal(ObjectOutput objectOutput) throws IOException {
        this.time.writeExternal(objectOutput);
        this.offset.writeExternal(objectOutput);
    }
    
    static OffsetTime readExternal(ObjectInput objectInput) throws IOException, ClassNotFoundException {
        return of(LocalTime.readExternal(objectInput), ZoneOffset.readExternal(objectInput));
    }
    
    static {
        MIN = LocalTime.MIN.atOffset(ZoneOffset.MAX);
        MAX = LocalTime.MAX.atOffset(ZoneOffset.MIN);
    }
}
